/*****************************************************************
 * File:		$RCSfile: AgmII.h,v $
 * Date:		$Date: 2005/05/23 09:30:08 $
 * Author:		$Author: haikun $
 * Revision:	$Revision: 1.9 $
 * License:		See file LICENSE
 *******************************************************************/

#ifndef __AGMII_H
#define __AGMII_H

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <limits.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <usb.h>
#include <pcsclite.h>

#include "agmdebug.h"
#include "checksum.h"
#include "LinuxDefines.h"
#include "atr.h"
#include "T1Protocol.h"
#include "tdssccom.h"
#include "ifdhandler.h"

/*****************************************************************************
*
*****************************************************************************/
#define USB_INEP    0			  //endpoint 0x81
#define USB_OUTEP   1             //endpoint 0x02

/*****************************************************************************
* IOCTL
*****************************************************************************/

/*****************************************************************************
*
*****************************************************************************/
#define LEFT_NIBBLE(c) ((c & 0xF0) >> 4)
#define RIGHT_NIBBLE(c) (c & 0x0F)
#define PRINT_CHAR(c) LEFT_NIBBLE(c), RIGHT_NIBBLE(c)

#define MIN(a, b) (a > b ? b : a)
#define MAX(a, b) (a > b ? a : b)

/*****************************************************************************
*
*****************************************************************************/
/* return codes */
#define AGM_OK							0
#define AGM_ERROR_DATA_INCOMPLETE		-1
#define AGM_ERROR_IOCTL_RESPONSE		-2
#define AGM_ERROR_IOCTL_EVENT			-3
#define AGM_ERROR_READ					-4
#define AGM_ERROR_WRITE					-5
#define AGM_ERROR_TIMEOUT				-6
#define AGM_ERROR_COMMAND				-7
#define AGM_ERROR_CHECKSUM				-8
#define AGM_ERROR_RESPONSE_INCORRECT	-9
#define AGM_ERROR_OVERHEATING			-10
#define AGM_ERROR_ATR					-11
#define AGM_ERROR_PARITY				-12
#define AGM_ERROR_FIFO_OVERRUN			-13
#define AGM_ERROR_FRAMING				-14
#define AGM_ERROR_EARLY_ANSWER			-15
#define AGM_ERROR_RESEND_COMMAND		-16

#define AGM_READER_PID_ERROR				-110
#define AGM_READER_CNT_ERROR				-111
#define AGM_READER_TRUNC_ERROR				-112
#define AGM_READER_LEN_ERROR				-113
#define AGM_READER_UNKNOWN_CMD_ERROR		-114
#define AGM_READER_TIMEOUT_ERROR			-115
#define AGM_READER_CS_ERROR					-116
#define AGM_READER_INVALID_PARAM_ERROR		-117
#define AGM_READER_CMD_FAILED_ERROR			-118
#define AGM_READER_NO_CARD_ERROR			-119
#define AGM_READER_CARD_NOT_POWERED_ERROR	-120
#define AGM_READER_COMM_ERROR				-121
#define AGM_READER_EXTRA_WAITING_TIME		-122

#define AGM_READER_NOT_CPU_CARD				-123
#define AGM_READER_NO_MATCHING_PARAMS		-124
#define AGM_READER_CARD_REJECTED			-125

#define AGM_READER_INVALID_STATUS_BYTE		-126
#define AGM_READER_RETRY_FAILED				-127

/*****************************************************************************
*
*****************************************************************************/
#define EVENT_BUFFER_SIZE       16
#define RESPONSE_BUFFER_SIZE    4096

#define MAX_SOCKET_NUM          1
#define MAX_READER_NUM          2
/*****************************************************************************
*
*****************************************************************************/
struct card_params
{
	uchar protocol;
	uchar N;
	uchar CWT[3];
	uchar BWT[3];
	uchar A;
	uchar B;
	uchar freq;
	float fTod;
};

typedef struct
{
	int status;						  /* 0 - absent, 1 - inserted, 2 - powered on */

	ATR atr;

	int cwt;						  /* max time in milliseconds to wait between characters */
	int bwt;						  /* max time in milliseconds to wait to first character (in T=0, equals to cwt) */

	t1_params t1Params;
	struct card_params cardParams;

	char activeProtocol;
} card;

typedef struct
{
	HANDLE handle;
	unsigned char baud;
	unsigned char bits;
	int stopbits;
	char parity;
	long blocktime;

} ioport;

#define BUS_DEVICE_STRSIZE 32

typedef struct
{
	usb_dev_handle *handle;
	uint8_t input_endpoint;
	uint8_t output_endpoint;
	uint8_t interface;
	struct usb_device *dev;
	char bus_device[BUS_DEVICE_STRSIZE];

	char response[RESPONSE_BUFFER_SIZE];
	unsigned int curPos, lastPos;

	char stopReading;
} usbDevice;

typedef struct
{
	usbDevice io;
	//int                 readerStarted; /* 1 - started, 0 - not yet */
	//char                commandCounter;
	card card;

	pthread_mutex_t semaphore;
    bool    ifsAdjust;
    int     ReadTimeout;                /* Reader ReadTimeout value in milliseconds */
    unsigned char CurrentDi;            /* Current reader used Di value , default to 1 */

} reader;
//*********************************************************************
//*
//* extern functions
//*
//*********************************************************************
extern void cur_buf_init(cur_buf *, void *, size_t);
extern void cur_buf_set(cur_buf *, void *, size_t);
extern int cur_buf_get(cur_buf *, void *, size_t);
extern int cur_buf_put(cur_buf *, const void *, size_t);
extern int cur_buf_putc(cur_buf *, int);
extern unsigned int cur_buf_avail(cur_buf *);
extern void *cur_buf_head(cur_buf *);

extern int t1_init(t1_params * t1);
extern int t1_set_param(t1_params * t1, int type, long value);

extern int GetStatus(reader * globalData, unsigned char *response, int *len);
extern int CardPowerOn(reader * globalData);
extern int CardPowerOff(reader * globalData);
extern int DataPackageCheck(unsigned char *tempBuffer);

extern bool OpenUSB(reader * allReaders, reader * globalData);
extern bool CloseUSB(reader * globalData);
extern int WriteUSB(reader * globalData, unsigned char *data, int timeOut);
extern int ReadUSB(reader * globalData, unsigned char *buf, int timeOut);

void CleanReadBufferUSB(reader * globalData);

int ParseATR(reader * globalData, char data[], int len);

extern int AGMTransmitT0(reader * globalData, PUCHAR TxBuffer, DWORD TxLength,
								 PUCHAR RxBuffer, PDWORD RxLength);

extern int AGMTransmitT1(reader * globalData, PUCHAR TxBuffer, DWORD TxLength,
								 PUCHAR RxBuffer, PDWORD RxLength);

extern unsigned long T0CmdParsing(unsigned char *cmd, unsigned int cmd_len,
											 unsigned int *exp_len);

extern void lock_mutex(reader * globalData);
extern void unlock_mutex(reader * globalData);

// ATR parse functions
extern uchar GetT1CWI(ATR * atr);

extern uchar GetT1BWI(ATR * atr);

extern uchar GetFi(ATR * atr);

extern uchar GetDi(ATR * atr);

extern uchar GetT1IFSC(ATR * atr);

extern uchar GetT1EDC(ATR * atr);

extern int GetCurrentMode (ATR* atr);

extern uchar GetSpecificModeProtocol (ATR* atr);

extern uchar GetFirstOfferedProtocol (ATR* atr);

extern float GetEtu(ATR * atr);

extern int GetCWT(ATR * atr);

extern int GetBWT(ATR * atr);

extern int GetT0CWT(ATR * atr);

extern uchar GetN (ATR* atr);

#endif
