/*****************************************************************
 * File:		$RCSfile: T1Protocol.c,v $
 * Date:		$Date: 2005/05/23 04:45:08 $
 * Author:		$Author: haikun $
 * Revision:	$Revision: 1.13 $
 *
 * Purpose:		This provides reader specific T1 protocol level calls
 * 				for the ArgosMini II (USB) reader of Todos Data System AB.
 * License:		See file LICENSE
 *******************************************************************/

#include "AgmII.h"

static unsigned int
t1_build(t1_params * t1, unsigned char *block,
			unsigned char dad, unsigned char pcb, cur_buf * bp, size_t * lenp);

static unsigned int t1_rebuild(t1_params *t1, unsigned char *block);

static unsigned t1_block_type(unsigned char pcb);

static unsigned int t1_seq(unsigned char pcb);

static void t1_set_defaults(t1_params * t1);

static unsigned int t1_compute_checksum(t1_params *, unsigned char *, size_t);

static void t1_set_checksum(t1_params *, int);

/*-----------------2005-5-10 13:40------------------
 * T1 transmit function
 * --------------------------------------------------*/
int AGMTransmitT1(reader * globalData, PUCHAR TxBuffer, DWORD TxLength,
						PUCHAR RxBuffer, PDWORD RxLength)
{
	unsigned char SetT1Str[ARGMINIUSB_FIFO_LEN] =
		{ CommandID, AGMU_T1_Data_length, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
	// sdata is block we have added NAD,PCB,LEN & EDC to the TxBuffer
	unsigned char sdata[T1_BUFFER_SIZE];
    unsigned char rdata[T1_BUFFER_SIZE];
    cur_buf sbuf, rbuf, tbuf;//,ifsbuf;
	unsigned char dad = 0;
	unsigned int slen;
	unsigned char SendMoreData = 1;
	unsigned char ReadMoreData = 1;
	unsigned int bytesToSend = 0;
	unsigned int bytesToRead = 0;
	unsigned char tempBuffer[ARGMINIUSB_FIFO_LEN] = { 0x00 };
	size_t last_send;
	int retVal = 0;
	int TotalReplyLength;
	int ReadTimeout;
    unsigned char   pcb,sblk[5];
    unsigned int retries, resyncs,sent_length=0;
    unsigned int numtimeout = 0;
	int wtx = 0;

 //   unsigned char ifsBuffer[] = {0xFF};

    AgmLog1(AGM_LOG_DEBUG, "AGMTransmitT1 enter!\n");

    globalData->card.t1Params.state = SENDING;
    retries = globalData->card.t1Params.retries;
    resyncs = 3;

    cur_buf_set(&sbuf, (void *) TxBuffer, TxLength);
    cur_buf_set(&rbuf, (void *) rdata, T1_BUFFER_SIZE);
//    cur_buf_set(&ifsbuf, (void *) ifsBuffer, sizeof(ifsBuffer));

    *RxLength = 0;
    AgmLog2(AGM_LOG_DEBUG, "t1Params.ifsc = %d", globalData->card.t1Params.ifsc);
    AgmLog2(AGM_LOG_DEBUG, "t1Params.ifsd = %d", globalData->card.t1Params.ifsd);
    AgmLog2(AGM_LOG_DEBUG, "t1Params.retries = %d", globalData->card.t1Params.retries);
    /*-----------------2005-5-10 13:58------------------
    * Send IFS adjustment request to card after reset,
    * according to EMV requirement
    * --------------------------------------------------*/
//    if( globalData->card.t1Params.ifsc == globalData->card.t1Params.ifsd )
    {
       globalData->ifsAdjust = TRUE;
       slen = t1_build(&globalData->card.t1Params, sdata, dad, T1_I_BLOCK, &sbuf,
                        &last_send);
    }


    while(retries >= 0)
    {
       // cur_buf_init(&rbuf, (void *) RxBuffer, T1_BUFFER_SIZE);
        cur_buf_set(&rbuf, (void *) rdata, T1_BUFFER_SIZE);
/*
        if( !globalData->ifsAdjust )
        {

            slen =
                t1_build(&globalData->card.t1Params, sdata, dad, T1_S_BLOCK | T1_S_IFS, &ifsbuf,
                        &last_send);
            globalData->ifsAdjust = TRUE;
        }
 */
        AgmLog2(AGM_LOG_DEBUG, "slen = %d", slen);
        AgmLogXxd(AGM_LOG_DEBUG, "sdata - the T1 block we had built: ", sdata, slen);

        retries--;

        AgmLog2(AGM_LOG_DEBUG, "retries-- = %d", retries);
        // send block
        do
        {
            int i = 0;
            memset( tempBuffer, 0 , sizeof(tempBuffer));
            while ((i < 6) && (bytesToSend < slen))
            {
                tempBuffer[i + 2] = sdata[bytesToSend];
                i++;
                bytesToSend++;
            }
            tempBuffer[0] = 0x01;
            tempBuffer[1] = i;

   //         AgmLogXxd(AGM_LOG_DEBUG, "WritePackage: ", tempBuffer,
   //                     ARGMINIUSB_FIFO_LEN);
            retVal = WriteUSB(globalData, tempBuffer, AGM_DEFAULT_TIMEOUT);

            if (retVal <= 0)
            {
                AgmLog1(AGM_LOG_DEBUG, "=====Write T0 data failed!=============\n");
                return AGM_ERROR_WRITE;
            }

//            AgmLog3(AGM_LOG_DEBUG, "bytesToSend = %d, TxLength = %d", bytesToSend,
 //                   TxLength);
            if (bytesToSend == slen)
            {
                SendMoreData = 0;
            }
        }
        while (SendMoreData);

        // restore bytesToSend to 0
        bytesToSend = 0;
        SendMoreData = 1;

        // build setT1 string
        SetT1Str[2] = slen;
        SetT1Str[3] = GetT1EDC(&globalData->card.atr);
        // Waiting Time Extension default to 0
		SetT1Str[4] = wtx;
//		AgmLog2(AGM_LOG_DEBUG, "SetT1Str[4]:%x\n", SetT1Str[4]);
        SetT1Str[5] = globalData->CurrentDi;    //GetDi ( &globalData->card.atr );
        SetT1Str[6] = GetT1CWI(&globalData->card.atr);
        SetT1Str[7] = GetT1BWI(&globalData->card.atr);

        //send SetT1str
        AgmLogXxd(AGM_LOG_DEBUG, "Write SetT1Str: ", SetT1Str, ARGMINIUSB_FIFO_LEN);
        retVal = WriteUSB(globalData, SetT1Str, AGM_DEFAULT_TIMEOUT);
        if (retVal <= 0)
        {
            AgmLog1(AGM_LOG_DEBUG, "=====Write SetT1Str failed!=============\n");
            return AGM_ERROR_WRITE;
        }

        TotalReplyLength = 0;
		if( wtx )
			ReadTimeout = 2*(GetBWT(&globalData->card.atr)*wtx+T1_BUFFER_SIZE*GetCWT(&globalData->card.atr))+AGM_DEFAULT_TIMEOUT;
		else
			ReadTimeout = 2*(GetBWT(&globalData->card.atr)*1+T1_BUFFER_SIZE*GetCWT(&globalData->card.atr))+AGM_DEFAULT_TIMEOUT;

		wtx = 0;
//		AgmLog2(AGM_LOG_DEBUG, "ReadTimeout:%d\n", ReadTimeout);
        // restore bytesToRead to 0
        bytesToRead = 0;
        ReadMoreData = 1;
        do
        {
            memset( tempBuffer, 0 , sizeof(tempBuffer));

            retVal = ReadUSB(globalData, tempBuffer, ReadTimeout);
            AgmLogXxd(AGM_LOG_DEBUG, "Read package: ", tempBuffer, ARGMINIUSB_FIFO_LEN);
            if (retVal <= 0)
            {
                AgmLog1(AGM_LOG_DEBUG,"=====Read first package failed!=============");
                /* do same for BWT and CWT timeout , should be difference later according to EMV*/
                if ( retries == 0 )
                {
                    retries = 1;
                }
                // Send R_Block to card
                AgmLog2(AGM_LOG_DEBUG, "numtimeout = %d", numtimeout);
                if( numtimeout > 4 )
                    return AGM_ERROR_TIMEOUT;
                if( numtimeout < 2 )
                {
                    numtimeout++;
                    slen = t1_build(&globalData->card.t1Params, sdata,dad, T1_R_BLOCK, NULL, NULL);
                    continue;
                }
                else
                {
                    numtimeout++;
                    goto resync;
                }
                return AGM_ERROR_READ;
            }
            if (CommandID == tempBuffer[0])
            {
                if (AGMU_CardStatus_IND == tempBuffer[1])
                {
                    if (0 == tempBuffer[2])
                    {
                        // No card;
                        AgmLog1( AGM_LOG_DEBUG, "=====No card in reader!=============");
                        return AGM_READER_NO_CARD_ERROR;
                    }
                }
                /*-----------------2005-5-23 10:52------------------
                 * add BWT timeout handler according to
                 *
                 * ISO7816-3 rules 6.4,7.1,7.4.2 and 7.4.3
                 *
                 * --------------------------------------------------*/
                else if( AGMU_T1timeout_IND == tempBuffer[1])
                {
                    /* do same for BWT and CWT timeout , should be difference later according to EMV*/
                    if ( retries == 0 )
                    {
                        retries = 1;
                    }
                    // Send R_Block to card
                    AgmLog2(AGM_LOG_DEBUG, "numtimeout = %d", numtimeout);
                    if( numtimeout > 4 )
                        return AGM_ERROR_TIMEOUT;
                    if( numtimeout < 2 )
                    {
                        numtimeout++;
                        slen = t1_build(&globalData->card.t1Params, sdata,dad, T1_R_BLOCK, NULL, NULL);
                        continue;
                    }
                    else
                    {
                        numtimeout++;
                        goto resync;
                    }
                }
                else
                {
                    AgmLog1(AGM_LOG_DEBUG,
                            "=====Unknown reader error!=============");
                    return AGM_ERROR_READ;
                }
            }
            else
            {
                // Check for card error or removal.
                if (bytesToRead == 0)
                {
                    // First package.
                    //
                    ReadTimeout = 100;
                    if (1 == tempBuffer[1])
                    {
                        // Error in firmware
                        // Probebly a timeout.
                        return AGM_ERROR_TIMEOUT;
                    }
                }
                // Todo Check length
                if (tempBuffer[1] <= ARGMINIUSB_FIFO_LEN - 2)
                {
                    memcpy(rbuf.base + rbuf.head, &tempBuffer[2], tempBuffer[1]);
                    bytesToRead += tempBuffer[1];
                    rbuf.head += tempBuffer[1];
                }
                else
                {
                    AgmLog2(AGM_LOG_DEBUG, "Data package length error! Length = %d",
                            tempBuffer[1]);
                    return AGM_ERROR_READ;
                }
                if (0 == TotalReplyLength)
                {
                    //
                    // The third byte contains the length of the data in the packet.
                    // Check if we have three bytes
                    //
                    if (bytesToRead >= 3)
                    {
                        //
                        // The third byte contains the length of the data in the packet
                        // and we additinally want to have the EDC bytes which
                        // is one for LRC and 2 for CRC
                        //
                        TotalReplyLength = rbuf.base[2] + 3 +
                            GetT1EDC(&globalData->card.atr);

                        AgmLog2(AGM_LOG_DEBUG, "Read TotalReplyLength = %d\n",
                                TotalReplyLength);
                    }
                }
                if (TotalReplyLength == bytesToRead)
                {
                    // A T1 package is recived!
                    ReadMoreData = 0;
                    AgmLogXxd(AGM_LOG_DEBUG, "Read T1 block :", rbuf.base,
                                bytesToRead);
                }
            }
            AgmLog3(AGM_LOG_DEBUG, "BytesToRead = %d, BytesRead = %d",
                    TotalReplyLength, bytesToRead);
        }
        while (ReadMoreData);

        // Check incoming block
        pcb = rbuf.base[PCB];
        AgmLog2(AGM_LOG_DEBUG, "read block pcb = %x", pcb);
        switch (t1_block_type(pcb)) {
        case T1_R_BLOCK :
            if ((rbuf.base[LEN] != 0x00)    /* length != 0x00 (illegal) */
				|| (pcb & 0x20)			/* b6 of pcb is set */
			   )
			{
                AgmLog1(AGM_LOG_DEBUG,"R-Block required");
				/* ISO 7816-3 Rule 7.4.2 */
				if (retries == 0)
					goto resync;

				/* ISO 7816-3 Rule 7.2 */
                if (T1_R_BLOCK == t1_block_type(globalData->card.t1Params.previous_block[1]))
				{
                    AgmLog1(AGM_LOG_DEBUG,"Rule 7.2");
                    slen = t1_rebuild(&globalData->card.t1Params, sdata);
					continue;
				}

                slen = t1_build(&globalData->card.t1Params, sdata,
						dad, T1_R_BLOCK | T1_OTHER_ERROR,
						NULL, NULL);
				continue;
            }

            if (((t1_seq(pcb) != globalData->card.t1Params.ns)    /* wrong sequence number & no bit more */
                    && ! globalData->card.t1Params.more)
			   )
			{
                AgmLog4(AGM_LOG_DEBUG,"received: %d, expected: %d, more: %d",
                    t1_seq(pcb), globalData->card.t1Params.ns, globalData->card.t1Params.more);

				/* ISO 7816-3 Rule 7.2 */
                if (T1_R_BLOCK == t1_block_type(globalData->card.t1Params.previous_block[PCB]))
				{
                    AgmLog1(AGM_LOG_DEBUG,"Rule 7.2");
                    slen = t1_rebuild(&globalData->card.t1Params, sdata);
					continue;
				}

                AgmLog1(AGM_LOG_DEBUG,"R-Block required");
				/* ISO 7816-3 Rule 7.4.2 */
				if (retries == 0)
					goto resync;
                slen = t1_build(&globalData->card.t1Params, sdata,
						dad, T1_R_BLOCK | T1_OTHER_ERROR,
						NULL, NULL);
				continue;
            }

            if (globalData->card.t1Params.state == RECEIVING) {
				/* ISO 7816-3 Rule 7.2 */
                if (T1_R_BLOCK == t1_block_type(globalData->card.t1Params.previous_block[1]))
				{
                    AgmLog1(AGM_LOG_DEBUG,"Rule 7.2");
                    slen = t1_rebuild(&globalData->card.t1Params, sdata);
					continue;
				}

                slen = t1_build(&globalData->card.t1Params, sdata,
						dad, T1_R_BLOCK,
						NULL, NULL);
				break;
            }

            /* If the card terminal requests the next
			 * sequence number, it received the previous
			 * block successfully */
            if (t1_seq(pcb) != globalData->card.t1Params.ns) {
                cur_buf_get(&sbuf, NULL, last_send);
				sent_length += last_send;
//                last_send = 0;
                globalData->card.t1Params.ns ^= 1;
			}

			/* If there's no data available, the ICC
			 * shouldn't be asking for more */
            if (cur_buf_avail(&sbuf) == 0)
				goto resync;

            if( T1_IS_ERROR(pcb) == T1_OTHER_ERROR )
            {
                AgmLog1(AGM_LOG_DEBUG,"R_BLOCK other error");
//                if (retries == 0)
                    goto resync;
            }

            /* Check if there is M-bit set in previous I_BLOCK*/
            if(T1_I_BLOCK == t1_block_type(globalData->card.t1Params.previous_block[PCB]))
            {
                AgmLog1(AGM_LOG_DEBUG,"T1_MORE_BLOCKS in previous I_BLOCK");
                if( (globalData->card.t1Params.previous_block[PCB] & T1_MORE_BLOCKS) == T1_MORE_BLOCKS )
                {
               //     if( retries == 0 )
              //          retries += 1;
                    AgmLog2(AGM_LOG_DEBUG, "retries = %d", retries);

                    AgmLog3(AGM_LOG_DEBUG, "last_send = %d. sbuf.size = %d", last_send,sbuf.size);
                    if( retries > 0 )
                    {
                        slen = t1_build(&globalData->card.t1Params, sdata, dad, T1_I_BLOCK,&sbuf, &last_send);
                        continue;
                    }
                    retries++;
                    slen = t1_build(&globalData->card.t1Params, sdata, dad, T1_I_BLOCK,&sbuf, &last_send);
                    continue;
//                    break;
                }
            }

  //          slen = t1_build(&globalData->card.t1Params, sdata, dad, T1_I_BLOCK,
  //                  &sbuf, &last_send);
            break;
 //           continue;

        case T1_I_BLOCK :
            /* The first I-block sent by the ICC indicates
            * the last block we sent was received successfully. */
            if (globalData->card.t1Params.state == SENDING) {
                cur_buf_get(&sbuf, NULL, last_send);
                last_send = 0;
                globalData->card.t1Params.ns ^= 1;
            }

            globalData->card.t1Params.state = RECEIVING;
            /* If the block sent by the card doesn't match
            * what we expected it to send, reply with
            * an R block */
            if (t1_seq(pcb) != globalData->card.t1Params.nr) {
                AgmLog2(AGM_LOG_DEBUG, "retries = %d", retries);
                if( retries > 0 )
                {
                    memcpy( RxBuffer+(*RxLength),rbuf.base+3, rbuf.base[LEN]);
                    *RxLength += rbuf.base[LEN];
                    slen = t1_build(&globalData->card.t1Params, sdata, dad,
                        T1_R_BLOCK | T1_OTHER_ERROR,
                        NULL, NULL);
                    continue;
                }
                memcpy( RxBuffer+(*RxLength),rbuf.base+3, rbuf.base[LEN]);
                *RxLength += rbuf.base[LEN];
                break;
            }


            globalData->card.t1Params.nr ^= 1;

//            *RxLength = cur_buf_put(&rbuf, rbuf.base + 3, rbuf.base[LEN]) ;
            memcpy( RxBuffer+(*RxLength),rbuf.base+3, rbuf.base[LEN]);
            *RxLength += rbuf.base[LEN];

            if ( *RxLength < 0)
            {
                AgmLog1(AGM_LOG_DEBUG, "cur_buf_put got error with read data");
                globalData->card.t1Params.state = DEAD;
                return AGM_ERROR_READ;
            }

            if ((pcb & T1_MORE_BLOCKS) == 0)
            {
                AgmLog2(AGM_LOG_DEBUG, "BytesToRead = %d", *RxLength);
                AgmLogXxd(AGM_LOG_DEBUG, "RxBuffer block :", RxBuffer,*RxLength);
                AgmLog1(AGM_LOG_DEBUG, "Read Data success!");
                return AGM_OK;
            }

            slen = t1_build(&globalData->card.t1Params, sdata, dad, T1_R_BLOCK, NULL, NULL);
            retries = globalData->card.t1Params.retries;
//          break;
            continue;
        case T1_S_BLOCK :
            if (T1_S_IS_RESPONSE(pcb) && globalData->card.t1Params.state == RESYNCH) {
				/* ISO 7816-3 Rule 6.2 */
                AgmLog1(AGM_LOG_DEBUG,"S-Block answer received");
				/* ISO 7816-3 Rule 6.3 */
                globalData->card.t1Params.state = SENDING;
                sent_length =0;
				last_send = 0;
				resyncs = 3;
                retries = globalData->card.t1Params.retries;
                cur_buf_init(&rbuf, (void *) RxBuffer, T1_BUFFER_SIZE);
                slen = t1_build(&globalData->card.t1Params, sdata, dad, T1_I_BLOCK,
                       &sbuf, &last_send);
				continue;
            }
            /*-----------------2005-5-16 9:20-------------------
             * Add S_REPONSE | T1_S_IFS handler
             * --------------------------------------------------*/
            if(T1_S_IS_RESPONSE(pcb) && ((pcb & T1_S_IFS) == T1_S_IFS))
            {
                /* ISO 7816-3 Rule 4*/
                AgmLog1(AGM_LOG_DEBUG,"S-Block IFS repsonse received");
                globalData->card.t1Params.ifsc = rbuf.base[DATA];
                globalData->card.t1Params.state = SENDING;
                sent_length =0;
				last_send = 0;
				resyncs = 3;
                retries = globalData->card.t1Params.retries;
                slen = t1_build(&globalData->card.t1Params, sdata, dad, T1_I_BLOCK,
						&sbuf, &last_send);
                continue;
            }
            if (T1_S_IS_RESPONSE(pcb))
			{
				/* ISO 7816-3 Rule 7.4.2 */
				if (retries == 0)
					goto resync;

				/* ISO 7816-3 Rule 7.2 */
                if (T1_R_BLOCK == t1_block_type(globalData->card.t1Params.previous_block[PCB]))
				{
                    slen = t1_rebuild(&globalData->card.t1Params, sdata);
					continue;
				}

                AgmLog1(AGM_LOG_DEBUG,"wrong response S-BLOCK received");
                slen = t1_build(&globalData->card.t1Params, sdata,
						dad, T1_R_BLOCK | T1_OTHER_ERROR,
						NULL, NULL);
                continue;
            }

            cur_buf_init(&tbuf, sblk, sizeof(sblk));
            AgmLog1(AGM_LOG_DEBUG,"S-Block request received");
            switch (T1_S_TYPE(pcb)) {
			case T1_S_RESYNC:
                if (rbuf.base[LEN] != 0)
				{
                    AgmLog2(AGM_LOG_DEBUG,"Wrong length: %d", rbuf.base[LEN]);
                    slen = t1_build(&globalData->card.t1Params, sdata, dad,
						T1_R_BLOCK | T1_OTHER_ERROR,
						NULL, NULL);
					continue;
				}

                AgmLog1(AGM_LOG_DEBUG,"Resync requested");
				/* the card is not allowed to send a resync. */
				goto resync;

			case T1_S_ABORT:
                if (rbuf.base[LEN] != 0)
				{
                    AgmLog2(AGM_LOG_DEBUG,"Wrong length: %d", rbuf.base[LEN]);
                    slen = t1_build(&globalData->card.t1Params, sdata, dad,
						T1_R_BLOCK | T1_OTHER_ERROR,
						NULL, NULL);
					continue;
				}

				/* ISO 7816-3 Rule 9 */
                AgmLog1(AGM_LOG_DEBUG,"abort requested");
				goto resync;

			case T1_S_IFS:
                if (rbuf.base[LEN] != 1)
				{
                    AgmLog2(AGM_LOG_DEBUG,"Wrong length: %d", rbuf.base[LEN]);
                    slen = t1_build(&globalData->card.t1Params, sdata, dad,
						T1_R_BLOCK | T1_OTHER_ERROR,
						NULL, NULL);
					continue;
				}

                AgmLog2(AGM_LOG_DEBUG,"CT sent S-block with ifs=%u", rbuf.base[DATA]);
                if (rbuf.base[DATA] == 0)
					goto resync;
                globalData->card.t1Params.ifsc = rbuf.base[DATA];
                cur_buf_putc(&tbuf, rbuf.base[DATA]);
				break;

			case T1_S_WTX:
                if (rbuf.base[LEN] != 1)
				{
                    AgmLog2(AGM_LOG_DEBUG,"Wrong length: %d", rbuf.base[LEN]);
                    slen = t1_build(&globalData->card.t1Params, sdata, dad,
						T1_R_BLOCK | T1_OTHER_ERROR,
						NULL, NULL);
					continue;
				}

                AgmLog2(AGM_LOG_DEBUG,"CT sent S-block with wtx=%u", rbuf.base[DATA]);
                globalData->card.t1Params.wtx = rbuf.base[DATA];
                cur_buf_putc(&tbuf, rbuf.base[DATA]);
				wtx = rbuf.base[DATA];
//				AgmLog2(AGM_LOG_DEBUG, "Card request wtx:%d\n",wtx);
				break;

			default:
                AgmLog2(AGM_LOG_DEBUG,"T=1: Unknown S block type 0x%02x", T1_S_TYPE(pcb));
				goto resync;
            }

            slen = t1_build(&globalData->card.t1Params, sdata, dad,
				T1_S_BLOCK | T1_S_RESPONSE | T1_S_TYPE(pcb),
                &tbuf, NULL);

            /* Everything went just splendid */
            retries = globalData->card.t1Params.retries;
            continue;
        }
resync:
		/* the number or resyncs is limited, too */
		/* ISO 7816-3 Rule 6.4 */
		if (resyncs == 0)
			goto error;

		/* ISO 7816-3 Rule 6 */
		resyncs--;
        globalData->card.t1Params.ns = 0;
        globalData->card.t1Params.nr = 0;
        slen = t1_build(&globalData->card.t1Params, sdata, dad, T1_S_BLOCK|T1_S_RESYNC, NULL,
				NULL);
        globalData->card.t1Params.state = RESYNCH;
        globalData->card.t1Params.more = FALSE;
        retries = 1;
		continue;
error:
        globalData->card.t1Params.state = DEAD;
        return -1;

    };
	return AGM_OK;
}


/*
 * Attach t1 protocol
 */
int t1_init(t1_params * t1)
{
	t1_set_defaults(t1);			  //Set default T=1 protocol parameters
	t1_set_param(t1, IFD_PROTOCOL_T1_CHECKSUM_LRC, 0);	//default using LRC
	t1_set_param(t1, IFD_PROTOCOL_T1_STATE, SENDING);
	t1_set_param(t1, IFD_PROTOCOL_T1_MORE, FALSE);

	return 0;
}

/*
 * Build/verify checksum
 */
static unsigned int
t1_compute_checksum(t1_params * t1, unsigned char *data, size_t len)
{
	return len + t1->checksum(data, len, data + len);
}

static void t1_set_checksum(t1_params * t1, int csum)
{
	switch (csum)
	{
	case IFD_PROTOCOL_T1_CHECKSUM_LRC:
		t1->rc_bytes = 1;
		t1->checksum = csum_lrc_compute;
		break;
	case IFD_PROTOCOL_T1_CHECKSUM_CRC:
		t1->rc_bytes = 2;
		t1->checksum = csum_crc_compute;
		break;
	}
}

/*
 * Set default T=1 protocol parameters
 */
static void t1_set_defaults(t1_params * t1)
{
	t1->retries = 3;
	/* This timeout is rather insane, but we need this right now
	 * to support cryptoflex keygen */
	t1->ifsc = 32;
    t1->ifsd = 255;
	t1->nr = 0;
	t1->ns = 0;
	t1->wtx = 0;
}

/*
 * Get/set parmaters for T1 protocol
 */
int t1_set_param(t1_params * t1, int type, long value)
{
	switch (type)
	{
	case IFD_PROTOCOL_T1_CHECKSUM_LRC:
	case IFD_PROTOCOL_T1_CHECKSUM_CRC:
		t1_set_checksum(t1, type);
		break;
	case IFD_PROTOCOL_T1_IFSC:
		t1->ifsc = value;
		break;
	case IFD_PROTOCOL_T1_IFSD:
		t1->ifsd = value;
		break;
	case IFD_PROTOCOL_T1_STATE:
		t1->state = value;
		break;
	case IFD_PROTOCOL_T1_MORE:
		t1->more = value;
		break;
	default:
		return -1;
	}

	return 0;
}

// block type
static unsigned t1_block_type(unsigned char pcb)
{
	switch (pcb & 0xC0)
	{
	case T1_R_BLOCK:
		return T1_R_BLOCK;
	case T1_S_BLOCK:
		return T1_S_BLOCK;
	default:
		return T1_I_BLOCK;
	}
}

static unsigned int
t1_seq(unsigned char pcb)
{
	switch (pcb & 0xC0) {
	case T1_R_BLOCK:
		return (pcb >> T1_R_SEQ_SHIFT) & 1;
	case T1_S_BLOCK:
		return 0;
	default:
		return (pcb >> T1_I_SEQ_SHIFT) & 1;
	}
}

// use the incoming parameters to build T1 transmit block
static unsigned int
t1_build(t1_params * t1, unsigned char *block,
			unsigned char dad, unsigned char pcb, cur_buf * bp, size_t * lenp)
{
	unsigned int len;
	char more = FALSE;

	AgmLog1(AGM_LOG_DEBUG, "t1_build enter!\n");
	AgmLog3(AGM_LOG_DEBUG, "dad = %x,pcb = %x\n", dad, pcb);

	// if bp is not null,return length is return bp->tail - bp->head;
	// else ,len is 0
	len = bp ? cur_buf_avail(bp) : 0;
	if (len > t1->ifsc)
	{
		//#define T1_MORE_BLOCKS      0x20
		pcb |= T1_MORE_BLOCKS;
		len = t1->ifsc;
		more = TRUE;
	}

	/* Add the sequence number */
	switch (t1_block_type(pcb))
	{
	case T1_R_BLOCK:
		pcb |= t1->nr << T1_R_SEQ_SHIFT;
		break;
	case T1_I_BLOCK:
		pcb |= t1->ns << T1_I_SEQ_SHIFT;
		t1->more = more;
		break;
	}

	block[0] = dad;
	block[1] = pcb;
	block[2] = len;

	if (len)
		memcpy(block + 3, cur_buf_head(bp), len);
	if (lenp)
		*lenp = len;

	len = t1_compute_checksum(t1, block, len + 3);

	/* memorize the last sent block */
	/* only 4 bytes since we are only interesed in R-blocks */
	memcpy(t1->previous_block, block, 4);

	return len;
}

static unsigned int
t1_rebuild(t1_params *t1, unsigned char *block)
{
	unsigned char pcb = t1 -> previous_block[1];

	/* copy the last sent block */
	if (T1_R_BLOCK == t1_block_type(pcb))
		memcpy(block, t1 -> previous_block, 4);
	else
    {
        return 0;
	}

	return 4;
}

void cur_buf_init(cur_buf * bp, void *mem, size_t len)
{
	memset(bp, 0, sizeof(*bp));
	bp->base = (unsigned char *) mem;
	bp->size = len;
}

void cur_buf_set(cur_buf * bp, void *mem, size_t len)
{
	cur_buf_init(bp, mem, len);
	bp->tail = len;

	AgmLog4(AGM_LOG_DEBUG, "cur_buf_set :head = %d,tail = %d,size = %d\n",
			  bp->head, bp->tail, bp->size);
}

int cur_buf_get(cur_buf * bp, void *mem, size_t len)
{
	if (len > bp->tail - bp->head)
		return -1;
	if (mem)
		memcpy(mem, bp->base + bp->head, len);
	bp->head += len;
	return len;
}

int cur_buf_put(cur_buf * bp, const void *mem, size_t len)
{
	if (len > bp->size - bp->tail)
	{
		bp->overrun = 1;
		return -1;
	}
	if (mem)
		memcpy(bp->base + bp->tail, mem, len);
	bp->tail += len;
	return len;
}

int cur_buf_putc(cur_buf * bp, int byte)
{
	unsigned char c = byte;

	return cur_buf_put(bp, &c, 1);
}

unsigned int cur_buf_avail(cur_buf * bp)
{
	return bp->tail - bp->head;
}

void *cur_buf_head(cur_buf * bp)
{
	return bp->base + bp->head;
}
