/*****************************************************************
 * File:		$RCSfile: tdssccom.c,v $
 * Date:		$Date: 2005/05/23 09:30:08 $
 * Author:		$Author: haikun $
 * Revision:	$Revision: 1.15 $
 *
 * Purpose:		This provides reader specific low-level calls
 * 				for the ArgosMini II (USB) reader of Todos Data System AB.
 * License:		See file LICENSE
 *******************************************************************/

#include <pcsclite.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "AgmII.h"

int GetT0CWT(ATR * atr);

static int fi[16] = { 372, 372, 558, 744, 1116, 1488, 1860, 1, 1, 512, 768, 1024, 1536, 2048, 1, 1 };
static int di[16] = { 1, 1, 2, 4, 8, 16, 32, 1, 12, 20, 1, 1, 1, 1, 1, 1 };
static long fs[16] = {4000000, 5000000, 6000000, 8000000, 12000000, 16000000, 20000000, 0,
                      0, 5000000, 7500000, 10000000, 15000000, 20000000, 0, 0};

/*****************************************************************************
*
*****************************************************************************/
void lock_mutex(reader * globalData)
{
	int retVal;

	retVal = pthread_mutex_lock(&(globalData->semaphore));
	if (retVal != 0)
	{
		AgmLog1(AGM_LOG_ERROR, "Error! - in pthread_mutex_lock\n");
	}
}

/*****************************************************************************
*
*****************************************************************************/
void unlock_mutex(reader * globalData)
{
	int retVal;
	retVal = pthread_mutex_unlock(&(globalData->semaphore));
	if (retVal != 0)
	{
		AgmLog1(AGM_LOG_ERROR, "Error! - in pthread_mutex_unlock\n");
	}
}


int DataPackageCheck(unsigned char *tempBuffer)
{
	if (DataID == tempBuffer[0])
	{
		if (tempBuffer[1] <= (ARGMINIUSB_FIFO_LEN - 2))
		{
			return 1;
		}
	}
	return 0;
}

/*****************************************************************************
*
*****************************************************************************/
int GetStatus(reader * globalData, unsigned char *response, int *len)
{

	int retVal = 0;
	unsigned char cmd[10], status[10];

	memset(cmd, 0x00, sizeof(cmd));
	memset(status, 0x00, sizeof(cmd));

	cmd[0] = 0x02;
	cmd[1] = AGMU_GetCardStatus;

	lock_mutex(globalData);
/*
 * Removing this read due to usb error.
 * reads after a timeout fails....
 * Double card status packages is handled in CardPowerOn()
 * retVal = ReadUSB( globalData, response, AGM_DEFAULT_TIMEOUT);
 */

	if (0 >= retVal)
	{
		retVal = WriteUSB(globalData, cmd, AGM_DEFAULT_TIMEOUT);
		retVal = ReadUSB(globalData, response, AGM_DEFAULT_TIMEOUT);
	}

	if (retVal > 0)
	{
//        AgmLogXxd(AGM_LOG_DEBUG, "Data read: ", response, 8);
		if (CommandID == response[0])
		{
			if (AGMU_CardStatus_IND == response[1])
			{
				if (0x01 == response[2])
				{
					globalData->card.status = IFD_ICC_PRESENT;
					retVal = 1;
				}
				else
				{
					globalData->card.status = IFD_ICC_NOT_PRESENT;
					retVal = 0;
				}
			}
			else
			{
				retVal = -1;
				AgmLog2(AGM_LOG_DEBUG, "Reader response code: %x\n", response[1]);
				AgmLog2(AGM_LOG_DEBUG, "Reader response code: %x\n", response[2]);
			}
		}
	}
	else
	{
        AgmLog2(AGM_LOG_DEBUG, "retVal: %d\n", retVal);
		retVal = -1;
	}

	unlock_mutex(globalData);

	return retVal;
}


/*****************************************************************************
*
*****************************************************************************/
int CardPowerOn(reader * globalData)
{

    int retVal, parseATR=AGM_ERROR_ATR;
    unsigned char cmdResetCard[ARGMINIUSB_FIFO_LEN] =
        { CommandID, AGMU_ResetCard, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
    unsigned char tempBuffer[ARGMINIUSB_FIFO_LEN];
    unsigned char atrBuffer[MAX_ATR_SIZE+1];
    int atrBufferLength = 0;

	int bytesRecieved = 0;
	int tempBufIndex = 0;
	int ReadTimeout = 11000;
	char FirstPackage = 1;
	BOOL UsePackage = TRUE;

	AgmLog1(AGM_LOG_DEBUG, "Enter CardPowerOn");
	memset(tempBuffer, 0, sizeof(tempBuffer));
    memset(atrBuffer, 0, sizeof(atrBuffer));

	retVal = WriteUSB(globalData, cmdResetCard, AGM_DEFAULT_TIMEOUT);
	if (retVal < 0)
	{
		AgmLog1(AGM_LOG_DEBUG, "Send Card Power command fail\n");
		retVal = AGM_ERROR_WRITE;
		return retVal;
	}

	while (1)
	{
		UsePackage = TRUE;
		retVal = ReadUSB(globalData, tempBuffer, ReadTimeout);
//        AgmLogXxd(AGM_LOG_DEBUG, "Data read: ", tempBuffer, 8);
		if (retVal > 0)
		{
			// Check if package is a AGMU_CardStatus_IND.
			if ( CommandID == tempBuffer[0] )
			{
				if ( AGMU_CardStatus_IND == tempBuffer[1] )
				{
					if ( 0 == tempBuffer[2] )
					{
						// No card in reader
						retVal = AGM_READER_NO_CARD_ERROR;
						break;
					}
					/*
					 * A AGMU_CardStatus_IND == true is expected
					 * and is silently ignored.
					 */
				}
				else
				{
					// Data error from reader.
					// Abort!
					retVal = AGM_READER_UNKNOWN_CMD_ERROR;
					break;
				}
			}
			else
			{
				if (!DataPackageCheck(tempBuffer))
				{
					retVal = AGM_ERROR_TIMEOUT;
					break;
				}

				if( FirstPackage )
				{
					if( 1 == tempBuffer[1] )
					{
						AgmLog1(AGM_LOG_DEBUG, "Reader sent timeout!");
						retVal = AGM_ERROR_TIMEOUT;
						break;
					}
					else
					{
						// Check first data!
                        // Must be 0x3B or 0x3F, others are valid
                        if( 0x3B != tempBuffer[2] && 0x3F != tempBuffer[2] )
						{
							AgmLog1(AGM_LOG_DEBUG, "Data error. Discard package!");
							UsePackage = FALSE;
						}
						FirstPackage = 0;
						ReadTimeout = 100;	  //AGM_DEFAULT_TIMEOUT;
					}
				}

				if( TRUE == UsePackage )
				{
					while (tempBufIndex < tempBuffer[1])
					{
                        atrBuffer[bytesRecieved] =
							tempBuffer[tempBufIndex + 2];
						bytesRecieved++;
						tempBufIndex++;
					}

					tempBufIndex = 0;
                    atrBufferLength = bytesRecieved;

                    if( AGM_OK == ParseATR( globalData, atrBuffer, atrBufferLength ) )
					{
						retVal = AGM_OK;
                        parseATR = AGM_OK;
                        continue;
					}
				}
			}
		}
		else
		{
            if( parseATR == AGM_OK)
            {
                retVal = AGM_OK;
                break;
            }
			retVal = AGM_ERROR_TIMEOUT;
			break;
		}
	}

//    AgmLogXxd(AGM_LOG_DEBUG, "atrBuffer block data: ", atrBuffer, atrBufferLength);

	/* Parse ATR */
	AgmLog2(AGM_LOG_DEBUG, "Read ATR length = %d\n",
			  globalData->card.atr.length);

    AgmLogXxd(AGM_LOG_DEBUG, "Data read: ", globalData->card.atr.data, globalData->card.atr.length);

	return retVal;
}

/*****************************************************************************
*
*****************************************************************************/
int CardPowerOff(reader * globalData)
{
	int retVal;
    unsigned char cmdPowerOffCard[ARGMINIUSB_FIFO_LEN] =
		{ CommandID, AGMU_CardPowerOff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

	retVal = WriteUSB(globalData, cmdPowerOffCard, AGM_DEFAULT_TIMEOUT);
	if (retVal <= 0)
	{
		AgmLog1(AGM_LOG_DEBUG, "Send Card Power down command fail");
		retVal = 0;
	}

	return retVal;

}

int ParseATR(reader * globalData, char data[], int len)
{
	uchar TDi;
	int pointer = 0, pn = 0, i;

    ATR *atr = &(globalData->card.atr);

    memset(atr, 0x00, sizeof(ATR));

	/* Store T0 and TS */
	if (len < 1)
		return AGM_ERROR_ATR;
	atr->TS = data[0];
	atr->data[pointer] = atr->TS;

	if (atr->TS == 0x03)
		atr->TS = 0x3F;

	if ((atr->TS != 0x3B) && (atr->TS != 0x3F))
		return AGM_ERROR_ATR;

	if (len < 2)
		return AGM_ERROR_ATR;

	atr->T0 = data[1];

	TDi = atr->T0;
	pointer = 1;
	atr->data[pointer] = atr->T0;

	/* Store number of historical bytes */
	atr->hbn = TDi & 0x0F;

	/* TCK is not present by default */
	(atr->TCK).present = 0;

//    AgmLog2(AGM_LOG_DEBUG, "TS = %x", TDi);
	/* Extract interface bytes */
	while (1)
	{
		/* Check TAi is present */
		if (TDi & 0x10)
		{
			pointer++;
			if (len < pointer)
			{
				return AGM_ERROR_ATR;
			}
			atr->ib[pn][ATR_INTERFACE_BYTE_TA].value = data[pointer];
			atr->ib[pn][ATR_INTERFACE_BYTE_TA].present = 1;
			atr->data[pointer] = atr->ib[pn][ATR_INTERFACE_BYTE_TA].value;
		}
		else
		{
			atr->ib[pn][ATR_INTERFACE_BYTE_TA].present = 0;
		}
		// Check TBi is present
		if (TDi & 0x20)
		{
			pointer++;
			if (len < pointer)
			{
				return AGM_ERROR_ATR;
			}
			atr->ib[pn][ATR_INTERFACE_BYTE_TB].value = data[pointer];
			atr->ib[pn][ATR_INTERFACE_BYTE_TB].present = 1;
            atr->data[pointer] = atr->ib[pn][ATR_INTERFACE_BYTE_TB].value;
		}
		else
		{
			atr->ib[pn][ATR_INTERFACE_BYTE_TB].present = 0;
		}


		/* Check TCi is present */
		if (TDi & 0x40)
		{
			pointer++;
			if (len < pointer)
			{
				return AGM_ERROR_ATR;
			}
			atr->ib[pn][ATR_INTERFACE_BYTE_TC].value = data[pointer];
			atr->ib[pn][ATR_INTERFACE_BYTE_TC].present = 1;
            atr->data[pointer] = atr->ib[pn][ATR_INTERFACE_BYTE_TC].value;
		}
		else
			atr->ib[pn][ATR_INTERFACE_BYTE_TC].present = 0;

		/* Check Tdi is present */
		if (TDi & 0x80)
		{
			pointer++;
			if (len < pointer)
			{
				return AGM_ERROR_ATR;
			}
			atr->ib[pn][ATR_INTERFACE_BYTE_TD].value = data[pointer];
			TDi = atr->ib[pn][ATR_INTERFACE_BYTE_TD].value;
			atr->ib[pn][ATR_INTERFACE_BYTE_TD].present = 1;
			atr->data[pointer] = atr->ib[pn][ATR_INTERFACE_BYTE_TD].value;
			(atr->TCK).present = ((TDi & 0x0F) != ATR_PROTOCOL_TYPE_T0);
			if (pn >= ATR_MAX_PROTOCOLS)
			{
				return AGM_ERROR_ATR;
			}
			pn++;
		}
		else
		{
			atr->ib[pn][ATR_INTERFACE_BYTE_TD].present = 0;
			break;
		}
	}

	/* Store number of protocols */
	atr->pn = pn + 1;

	/* Store historical bytes */
	for (i = 0; i < (atr->hbn); i++)
	{
		pointer++;
		if (len < pointer)
		{
			return AGM_ERROR_ATR;
		}


		atr->hb[i] = data[pointer];
		atr->data[pointer] = atr->hb[i];
 //       AgmLog2(AGM_LOG_DEBUG, "%c ", atr->hb[i]);
	}
 //   AgmLog1(AGM_LOG_DEBUG, "\n");

	/* store TCK */
	if ((atr->TCK).present)
	{
		pointer++;
		if (len < pointer)
		{
			return AGM_ERROR_ATR;
		}
		(atr->TCK).value = data[pointer];
		atr->data[pointer] = atr->TCK.value;
	}

	atr->length = pointer + 1;

//    AgmLogXxd(AGM_LOG_DEBUG, "atr block data: ", atr->data, atr->length);
    AgmLog1(AGM_LOG_DEBUG, "AGM_OK!");
	return AGM_OK;
}

/*****************************************************************************
*
*****************************************************************************/
uchar GetT1CWI(ATR * atr)
{
	int i;
	uchar tmp;
    for (i = 1; i < atr->pn; ++i)
	{
		/* check for the first occurance of T=1 in TDi */
		if (atr->ib[i][ATR_INTERFACE_BYTE_TD].present &&
			 (atr->ib[i][ATR_INTERFACE_BYTE_TD].value & 0x0F) ==
			 ATR_PROTOCOL_TYPE_T1)
		{
			/* return TA(i+1) if present */
			if (atr->ib[i + 1][ATR_INTERFACE_BYTE_TB].present)
			{
				tmp = (atr->ib[i + 1][ATR_INTERFACE_BYTE_TB].value & 0x0F);

				// Check max supported by reader.
				if( 5 < tmp )
				{
					return 5;
				}
			}
			else
			{
				return 5;			  /* default */
			}
		}
	}

	return 5;
}

/*****************************************************************************
*
*****************************************************************************/
uchar GetT1BWI(ATR * atr)
{
	int i;
	uchar tmp;

	for (i = 1; i < atr->pn; ++i)
	{
		/* check for the first occurance of T=1 in TDi */
		if (atr->ib[i][ATR_INTERFACE_BYTE_TD].present &&
			 (atr->ib[i][ATR_INTERFACE_BYTE_TD].value & 0x0F) ==
			 ATR_PROTOCOL_TYPE_T1)
		{
			/* return TA(i+1) if present */
			if (atr->ib[i + 1][ATR_INTERFACE_BYTE_TB].present)
			{
				tmp = ((atr->ib[i + 1][ATR_INTERFACE_BYTE_TB].value & 0xF0) >> 4);

				// Check max supported by reader.
				if( 4 < tmp )
				{
					return 4;
				}
			}
			else
			{
				return 4;			  /* default */
			}
		}
	}

	return 4;
}

/*****************************************************************************
*
*****************************************************************************/
uchar GetT1IFSC(ATR * atr)
{
	int i;

	for (i = 1; i < atr->pn; ++i)
	{
		/* check for the first occurance of T=1 in TDi */
		if (atr->ib[i][ATR_INTERFACE_BYTE_TD].present &&
			 (atr->ib[i][ATR_INTERFACE_BYTE_TD].value & 0x0F) ==
			 ATR_PROTOCOL_TYPE_T1)
		{
			/* return TA(i+1) if present */
			if (atr->ib[i + 1][ATR_INTERFACE_BYTE_TA].present)
				return atr->ib[i + 1][ATR_INTERFACE_BYTE_TA].value;
			else
				return 32;			  /* default */
		}
	}
	return 32;
}

/*****************************************************************************
*
*****************************************************************************/
uchar GetDi(ATR * atr)
{
	if (atr->ib[0][ATR_INTERFACE_BYTE_TA].present)
		return (atr->ib[0][ATR_INTERFACE_BYTE_TA].value & 0x0F);
	else
		return 1;					  /* 1 */
}

/*****************************************************************************
*
*****************************************************************************/
uchar GetFi(ATR * atr)
{
	if (atr->ib[0][ATR_INTERFACE_BYTE_TA].present)
		return (atr->ib[0][ATR_INTERFACE_BYTE_TA].value & 0xF0) >> 4;
	else
		return 1;					  /* 372 */
}

/*****************************************************************************
*
*****************************************************************************/
uchar GetN (ATR* atr) {
    if (atr->ib[0][ATR_INTERFACE_BYTE_TC].present)
        return (atr->ib[0][ATR_INTERFACE_BYTE_TC].value);
    else
        return 0;
}
/*****************************************************************************
*
*****************************************************************************/
uchar GetT1EDC(ATR * atr)
{
	int i;

	for (i = 1; i < atr->pn; ++i)
	{
		/* check for the first occurance of T=1 in TDi */
		if (atr->ib[i][ATR_INTERFACE_BYTE_TD].present &&
			 (atr->ib[i][ATR_INTERFACE_BYTE_TD].value & 0x0F) ==
			 ATR_PROTOCOL_TYPE_T1)
		{
			/* return TA(i+1) if present */
			if (atr->ib[i + 1][ATR_INTERFACE_BYTE_TC].present)
			{
				AgmLog2(AGM_LOG_DEBUG,
						  "atr->ib[i + 1][ATR_INTERFACE_BYTE_TC].value = %x\n",
						  atr->ib[i + 1][ATR_INTERFACE_BYTE_TC].value);
				return (atr->ib[i + 1][ATR_INTERFACE_BYTE_TC].value & 0x01) + 1;
			}
			else
				return 1;			  /* default - LRC */
		}
	}

	return 1;
}

/*****************************************************************************
*
*****************************************************************************/
int GetCurrentMode (ATR* atr) {
    /* TA2 */
    if (atr->ib[1][ATR_INTERFACE_BYTE_TA].present) {
        return SPECIFIC_MODE;
    }
    else
        return NEGOTIABLE_MODE;
}

/*****************************************************************************
*
*****************************************************************************/
uchar GetSpecificModeProtocol (ATR* atr) {
    if (GetCurrentMode(atr) == SPECIFIC_MODE)
        return ((atr->ib[1][ATR_INTERFACE_BYTE_TA].value) & 0x0F);
    else
        return UNKNOWN_PROTOCOL;
}

/*****************************************************************************
*
*****************************************************************************/
uchar GetFirstOfferedProtocol (ATR* atr) {
    /* only in NEGOTIABLE_MODE */
    if (atr->ib[0][ATR_INTERFACE_BYTE_TD].present)
        return (atr->ib[0][ATR_INTERFACE_BYTE_TD].value) & 0x0F;
    else
        return ATR_PROTOCOL_TYPE_T0;
}

/*****************************************************************************
 *
 *					T0CmdParsing
 *
 ****************************************************************************/
unsigned long T0CmdParsing(unsigned char *cmd, unsigned int cmd_len,
									unsigned int *exp_len)
{
	*exp_len = 0;

	/* Ref: 7816-4 Annex A */
	switch (cmd_len)
	{
	case 4:							  /* Case 1 */
		*exp_len = 2;				  /* SW1 and SW2 only */
		break;

	case 5:							  /* Case 2 */
		if (cmd[4] != 0)
			*exp_len = cmd[4] + 2;
		else
			*exp_len = 256 + 2;
		break;

	default:						  /* Case 3 */
		if (cmd_len > 5 && cmd_len == (unsigned int) (cmd[4] + 5))
			*exp_len = 2;			  /* SW1 and SW2 only */
		else
			return IFD_COMMUNICATION_ERROR;	/* situation not supported */
		break;
	}

	return IFD_SUCCESS;
}										  /* T0CmdParsing */


int AGMTransmitT0(reader * globalData, PUCHAR TxBuffer, DWORD TxLength,
						PUCHAR RxBuffer, PDWORD RxLength)
{
	AgmLog1(AGM_LOG_DEBUG, "=====Entering AGMTransmitT0=============\n");
	int retVal;
    unsigned char ReciveBuf[ARGMINIUSB_FIFO_LEN] = { 0x00 };
    unsigned char tempBuffer[ARGMINIUSB_FIFO_LEN] = { 0x00 };
	unsigned char SetT0Str[ARGMINIUSB_FIFO_LEN] =
        { CommandID, AGMU_T0_Data_length, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

	unsigned char SendMoreData = 1;
	unsigned int i = 0;

	unsigned int bytesToSend = 0;
	unsigned int bytesToRead = 0;
	unsigned int NumberOfBytesToRead = 0;
    int ReadTimeout;
    unsigned char * rbuf = RxBuffer;


	retVal = T0CmdParsing(TxBuffer, TxLength, &NumberOfBytesToRead);
    AgmLog2(AGM_LOG_DEBUG, "NumberOfBytesToRead value = %d\n", NumberOfBytesToRead);

	if (retVal != IFD_SUCCESS)
	{
		AgmLog1(AGM_LOG_DEBUG, "=====T0CmdParsing failed!=============\n");
		return IFD_COMMUNICATION_ERROR;
	}

    if( TxLength < 5 )
    {
        for(i=TxLength; i<5; i++)
        {
            TxBuffer[i] = 0;
        }
        TxLength = 5;
    }
    i=0;

    usleep(30000);

	tempBuffer[0] = DataID;

	do
	{
		while ((i < 6) && (bytesToSend < TxLength))
		{
			tempBuffer[i + 2] = TxBuffer[bytesToSend];
			i++;
			bytesToSend++;
		}
		tempBuffer[1] = i;
		i = 0;
        AgmLogXxd(AGM_LOG_DEBUG, "WritePackage", tempBuffer, ARGMINIUSB_FIFO_LEN);
		retVal = WriteUSB(globalData, tempBuffer, AGM_DEFAULT_TIMEOUT);

		if (retVal <= 0)
		{
			AgmLog1(AGM_LOG_DEBUG, "=====Write T0 data failed!=============\n");
			return AGM_ERROR_WRITE;
		}

		AgmLog3(AGM_LOG_DEBUG, "bytesToSend = %d, TxLength = %d", bytesToSend,
				  TxLength);
		if (bytesToSend == TxLength)
		{
			SendMoreData = 0;
		}
	}
	while (SendMoreData);

	SetT0Str[2] = (unsigned char) TxLength;
	SetT0Str[3] = globalData->CurrentDi;

    if (256 == NumberOfBytesToRead - 2)
	{
		SetT0Str[4] = 0xFF;
		SetT0Str[5] = 1;
	}
	else
	{
		SetT0Str[4] = (unsigned char) NumberOfBytesToRead - 2;
		SetT0Str[5] = 0;
	}

    AgmLogXxd(AGM_LOG_DEBUG, "Write SetT0Str", SetT0Str, ARGMINIUSB_FIFO_LEN);
	retVal = WriteUSB(globalData, SetT0Str, AGM_DEFAULT_TIMEOUT);
	if (retVal <= 0)
	{
		AgmLog1(AGM_LOG_DEBUG, "=====Write SetT0Str failed!=============\n");
		return AGM_ERROR_WRITE;
	}

	// read procedure byte
    AgmLog2(AGM_LOG_DEBUG, "GetT0CWT(&globalData->card.atr) value = %d\n", GetT0CWT(&globalData->card.atr));
    ReadTimeout = (bytesToSend + NumberOfBytesToRead )* GetT0CWT(&globalData->card.atr) + AGM_DEFAULT_TIMEOUT ;
    AgmLog2(AGM_LOG_DEBUG, "Readtimeout value = %d\n", ReadTimeout);

	do
	{
        retVal = ReadUSB(globalData, ReciveBuf, ReadTimeout);
        AgmLogXxd(AGM_LOG_DEBUG, "Read package ", ReciveBuf, ARGMINIUSB_FIFO_LEN);
		if (retVal <= 0)
		{
			AgmLog1(AGM_LOG_DEBUG,
					  "=====Read first package failed!=============\n");
			return AGM_ERROR_READ;
		}
		if (CommandID == ReciveBuf[0])
		{
			if (AGMU_CardStatus_IND == ReciveBuf[1])
			{
				if (0 == ReciveBuf[2])
				{
					// No card;
					AgmLog1(AGM_LOG_DEBUG,
							  "=====No card in reader!=============\n");
					return AGM_READER_NO_CARD_ERROR;
				}
			}
			else
			{
				AgmLog1(AGM_LOG_DEBUG,
						  "=====Unknown reader error!=============\n");
				return AGM_ERROR_READ;
			}
		}
		else
		{
			// Check for card error or removal.
			if (bytesToRead == 0)
			{
				// First package.
				//
                ReadTimeout = 100;
				if (1 == ReciveBuf[1])
				{
					// Error in firmware
					// Probebly a timeout.
					return AGM_ERROR_TIMEOUT;
				}
			}
			// Todo Check length
			if (ReciveBuf[1] <= ARGMINIUSB_FIFO_LEN - 2)
			{
                memcpy(rbuf, ReciveBuf+2, ReciveBuf[1]);
				bytesToRead += ReciveBuf[1];
                rbuf += ReciveBuf[1];
			}
			else
			{
				AgmLog2(AGM_LOG_DEBUG, "Data package length error! Length = %d\n",
						  ReciveBuf[1]);
				return AGM_ERROR_READ;
			}

			if (bytesToRead == 2)
			{
				break;
			}
		}
		AgmLog3(AGM_LOG_DEBUG, "BytesToRead = %d, BytesRead = %d",
				  NumberOfBytesToRead, bytesToRead);
	}
	while (bytesToRead < NumberOfBytesToRead);


	*RxLength = bytesToRead;
    AgmLogXxd(AGM_LOG_DEBUG, "Read package ", RxBuffer, *RxLength);

	return AGM_OK;
}

float GetEtu(ATR * atr)
{
	//declare
	uchar Fi, Di;
	float etu;

	Fi = GetFi(atr);
	Di = GetDi(atr);

	if((Fi <= 0)&&(Di <= 0))
	{
		AgmLog1(AGM_LOG_DEBUG, "The TA1 in the ATR was Wrong!\n");
	}

	//calculate the etu,-s
	if(fs[Fi])
	{
        etu = ((float)fi[Fi])/((float)di[Di])/3000000;
	}
	else
	{
        etu = 372.0/3000000;
	}

//	AgmLog2(AGM_LOG_DEBUG, "T1:etu = %f\n", etu);

	return etu;
}

/*
 * calculate the CWT value according to the CWI. *
 */
int GetCWT(ATR * atr)
{
	float CWT, etu;
	uchar CWI;

	etu = GetEtu(atr);
	CWI = GetT1CWI(atr);

	CWT = ((float)(1<<CWI)+11.0)*etu*1000+0.5;

	if((1-CWT)>0)
	{
		CWT = 1;
	}

//	AgmLog2(AGM_LOG_DEBUG, "T1:CWT = %f\n", CWT);

	return (int)CWT;
}
/*
 * calculate the BWT value according to the BWI and f.
 */
int GetBWT(ATR * atr)
{
	float BWT, etu;
	uchar BWI, Fi;

	etu = GetEtu(atr);
	Fi = GetFi(atr);
	BWI = GetT1BWI(atr);

	if(fs[Fi])
	{
        BWT = ((float)(1<<BWI)*960.0*372.0/3000000+11*etu)*1000;
	}
	else
	{
        BWT = ((float)(1<<BWI)*960.0*372.0/3000000+11*etu)*1000;
	}

//	AgmLog2(AGM_LOG_DEBUG, "T1:BWT = %f\n", BWT);

	return (int)BWT;
}

int GetT0CWT(ATR * atr)
{
	int iWI = 10;
	uchar Fi;
	float T0CWT;

	//get F from TA1
	Fi = GetFi(atr);
	if(Fi <= 0)
	{
		AgmLog1(AGM_LOG_DEBUG, "The TA1 in the ATR was Wrong!\n");
	}

	//get the WI from TC2 in the T0 ATR
	if (atr->ib[1][ATR_INTERFACE_BYTE_TC].present)
		iWI = atr->ib[1][ATR_INTERFACE_BYTE_TC].value;
	else
		iWI = 10;

	if(fs[Fi])
	{
        T0CWT = (float)(960*iWI*((float)fi[Fi])/3000000)*1000;
	}
	else
	{
        T0CWT = (float)(960*iWI*((float)fi[Fi])/3000000)*1000;
	}

	return (int)T0CWT;
}
